﻿using System;
using System.Collections.Generic;
using System.Text;
using WiimoteLib;
using System.Windows.Forms;
using System.Drawing;
using PointF = System.Drawing.PointF;
using ButtonState = WiimoteLib.ButtonState;
using System.Runtime.InteropServices;
using System.IO;
using System.Globalization;
using System.Diagnostics;       //process

using Ini;

//using System.Drawing.Drawing2D;

namespace WiimoteWhiteboard
{
    public class Controller
    {
        public const float CALIBRATION_MARGIN = .1f;
        const int SMOOTHING_BUFFER_SIZE = 50;

        public int iRightClickStatus = 0;   //add by gsyan
        public bool MoveCursor = false;     //modified by gsyan : change default value to false;
        bool PrevCursorEnabled;
        public bool CursorEnabled
        {
            get;
            set;
        }

        public struct CalibrationStateInfo
        {
            public int State;
            public System.Drawing.Point TargetPoint;
        }

        //add by gsyan
        public struct TogglesInfo 
        {
            public int type;
            public String mouse;
            public short key1;
            public short key2;
            public short key3;
            public short key4;
            public String ps_name;
            public String ps_path;
        }




        public event EventHandler<EventArgs<CalibrationStateInfo>> CalibrationStateChanged;
        public event EventHandler<EventArgs<WiimoteState>> StatusInfoChanged;

        public event EventHandler<EventArgs<System.Drawing.PointF>> ShowLoadingCircle;    //add by gsyan
        public event EventHandler<EventArgs> HideLoadingCircle;    //add by gsyan
        public event EventHandler<EventArgs<int>> ChangeBalloonTipText; //add by gsyan
        public event EventHandler<EventArgs<String>> ShowErrorMessage; //add by gsyan

        CalibrationStateInfo CalibrationState = new CalibrationStateInfo() { State = 0, TargetPoint = new System.Drawing.Point() };

        Warper warper = new Warper();
        PointF[] src = new PointF[4];
        PointF[] dst = new PointF[4];

        public PointF[] SourceLayout
        {
            get { return src; }
        }

        public PointF[] DestinationLayout
        {
            get { return dst; }
        }

        IRSensor lastWiiState = new IRSensor(); //helps with event firing
        ButtonState lastButtonState = new ButtonState();

        Object lockObj = new Object();

        PointF[] SmoothingBuffer = new PointF[SMOOTHING_BUFFER_SIZE];
        int SmoothingBufferIndex = 0;

        int clickCounter = 0;   //add by gsyan
        PointF clickPosition = new PointF(-1, -1);


        public int SmoothingAmount
        {
            get;
            set;
        }

        public bool SmoothingEnabled
        {
            get { return SmoothingAmount > 0; }
        }

        Rectangle ScreenSize;

        Wiimote wm = new Wiimote();
        InputController InputCont;

        Boolean isShowLoadingCircle = false;

        TogglesInfo[] togglesInfo=null;
        bool outsideScreenToggles = false;


        public Controller(Screen screen)
        {

            SmoothingAmount = 8;
            ScreenSize = screen.Bounds;
            InputCont = new InputController(ScreenSize);

            for (int i = 0; i < SMOOTHING_BUFFER_SIZE; i++)
                SmoothingBuffer[i] = new PointF();

            wm.WiimoteChanged += new EventHandler<WiimoteChangedEventArgs>(wm_WiimoteChanged);
            wm.WiimoteExtensionChanged += new EventHandler<WiimoteExtensionChangedEventArgs>(wm_WiimoteExtensionChanged);

            for (int i = 0; i < src.Length; i++)
            {
                src[i] = new PointF();
                dst[i] = new PointF();
            }

            CursorEnabled = LoadConfig();
            PrevCursorEnabled = CursorEnabled;

            IniFile iniFile = new IniFile(".\\OutsideScreenToggles.ini");
            if(iniFile.IniReadValue("global", "enabled") == "1") {
                outsideScreenToggles = true;
                togglesInfo = new TogglesInfo[12];
                for(int i=1; i<=12; i++) {
                    togglesInfo[i-1].type = Int32.Parse(iniFile.IniReadValue(i.ToString(), "type"));
                    togglesInfo[i-1].mouse = iniFile.IniReadValue(i.ToString(), "mouse");
                    togglesInfo[i - 1].key1 = Convert.ToInt16(iniFile.IniReadValue(i.ToString(), "key1"));
                    togglesInfo[i - 1].key2 = Convert.ToInt16(iniFile.IniReadValue(i.ToString(), "key2"));
                    togglesInfo[i - 1].key3 = Convert.ToInt16(iniFile.IniReadValue(i.ToString(), "key3"));
                    togglesInfo[i - 1].key4 = Convert.ToInt16(iniFile.IniReadValue(i.ToString(), "key4"));
                    togglesInfo[i-1].ps_name = iniFile.IniReadValue(i.ToString(), "program_name");
                    togglesInfo[i-1].ps_path = iniFile.IniReadValue(i.ToString(), "program_path");
                }
            }
        }

        public void Enable()
        {
            bool ok = true;
            try
            {
                wm.Connect();
            }
            catch (Exception e)
            {
                ok = false;
                Show_ErrorMessage(e.Message);
            }
            if (ok)
            {
                wm.SetReportType(InputReport.IRAccel, true);
                wm.GetStatus();
                wm.SetLEDs(true, false, false, false);
            }
        }

        public void Disable()
        {
            wm.Disconnect();
        }
        
        PointF GetSmoothedCursor(int amount)
        {
            int start = SmoothingBufferIndex - amount;
            if (start < 0)
                start = 0;
            PointF smoothed = new PointF(0, 0);
            int count = SmoothingBufferIndex - start;
            for (int i = start; i < SmoothingBufferIndex; i++)
            {
                smoothed.X += SmoothingBuffer[i % SMOOTHING_BUFFER_SIZE].X;
                smoothed.Y += SmoothingBuffer[i % SMOOTHING_BUFFER_SIZE].Y;
            }
            smoothed.X /= count;
            smoothed.Y /= count;
            return smoothed;
        }

        void wm_WiimoteExtensionChanged(object sender, WiimoteExtensionChangedEventArgs args)
        {
			if(!args.Inserted)
                wm.SetReportType(InputReport.IRAccel, true);
        }

        void wm_WiimoteChanged(object sender, WiimoteChangedEventArgs args)
        {
            lock (lockObj)
            {
                WiimoteLib.ButtonState bs = args.WiimoteState.ButtonState;

                //extract the wiimote state
                IRSensor ws = args.WiimoteState.IRState.IRSensors[0];

                int outsideToggle = 0;

                if (ws.Found) //move cursor
                {
                    PointF position = new PointF(ws.RawPosition.X, ws.RawPosition.Y);
                    PointF warpedPos = position;
                    warper.warp(position, ref warpedPos);

                    SmoothingBuffer[SmoothingBufferIndex % SMOOTHING_BUFFER_SIZE] = warpedPos;
                    SmoothingBufferIndex++;

                    if (clickCounter < 65534)
                    {
                        clickCounter++; //add by gsyan
                    }

                    if (!lastWiiState.Found) //mouse down
                    {
                        SmoothingBufferIndex = 0; //resets the count

                        clickCounter = 0;   //add by gsyan
                        clickPosition.X = ws.RawPosition.X;
                        clickPosition.Y = ws.RawPosition.Y;

                        if (CursorEnabled)
                        {
                            if (isShowLoadingCircle)
                            {
                                Hide_LoadingCircle(); //add by gsyan
                                isShowLoadingCircle = false;
                            }

                            if (warpedPos.X >= 0 && warpedPos.Y >= 0 && warpedPos.X<=ScreenSize.Width && warpedPos.Y <= ScreenSize.Height)
                            {   //--Inside Screen , general click
                                
                                InputCont.MoveMouse(warpedPos); //move the mouse to x,y

                                if (iRightClickStatus <1)                                    
                                    InputCont.LeftMouseDown(); //left click down
                                else
                                    InputCont.RightMouseDown();                                
                            }
                            else
                            { //--Outside Screen Toggle                                           
                                if (warpedPos.Y < 0)
                                { //top
                                    if(warpedPos.X >= 0 && warpedPos.X <= ScreenSize.Width)
                                    { //center
                                        if (warpedPos.X <= ScreenSize.Width / 2)
                                        {   //left
                                            outsideToggle = 2;
                                        }
                                        else
                                        {   //right
                                            outsideToggle = 3;
                                        }
                                    } else if(warpedPos.X > ScreenSize.Width) 
                                    {   //right corner
                                        outsideToggle = 4;
                                    } else 
                                    {   //left corner
                                        outsideToggle = 1;
                                    }
                                } else if (warpedPos.Y > ScreenSize.Height)
                                { //buttom 
                                    if (warpedPos.X >= 0 && warpedPos.X <= ScreenSize.Width)
                                    { //center
                                        //Change_BalloonTipText(1);
                                        //iRightClickStatus = 1;      //next to right click

                                        if (warpedPos.X <= ScreenSize.Width / 2)
                                        {   //left
                                            outsideToggle = 9;
                                        }
                                        else
                                        {   //right
                                            outsideToggle = 8;
                                        }
                                    } else if(warpedPos.X > ScreenSize.Width)
                                    {   //right corner
                                        outsideToggle = 7;
                                    }
                                    else
                                    {   //left corner
                                        outsideToggle = 10;
                                    }
                                }
                                else if (warpedPos.Y >= 0 && warpedPos.Y <= ScreenSize.Height)
                                {   //middle
                                    if(warpedPos.X > ScreenSize.Width) 
                                    {   //right
                                        if (warpedPos.Y <= ScreenSize.Height / 2)
                                        {   //upper
                                            //InputCont.SendKey(0x26);    //UP Arrow
                                            //Change_BalloonTipText(0x26);
                                            outsideToggle = 5;
                                        }
                                        else
                                        {   //lower
                                            //InputCont.SendKey(0x28);    //DOWN Arrow
                                            //Change_BalloonTipText(0x28);
                                            outsideToggle = 6;
                                        }
                                    } 
                                    else if (warpedPos.X < 0)
                                    {   //left 
                                        if (warpedPos.Y <= ScreenSize.Height / 2)
                                        {   //upper
                                            //InputCont.SendKey(0x26);    //UP Arrow
                                            //Change_BalloonTipText(0x26);
                                            outsideToggle = 12;
                                        }
                                        else
                                        {   //lower
                                            //InputCont.SendKey(0x28);    //DOWN Arrow
                                            //Change_BalloonTipText(0x28);
                                            outsideToggle = 11;
                                        }
                                    }
                                }
                                if (togglesInfo != null &&  outsideToggle > 0)
                                {
                                    short key;

                                    outsideToggle--;    //array index from 0 

                                    if (outsideScreenToggles && togglesInfo != null && togglesInfo[outsideToggle].type > 0)
                                    {
                                        switch (togglesInfo[outsideToggle].type)
                                        {
                                            case 1:
                                                if (togglesInfo[outsideToggle].mouse.ToUpper() == "RIGHT")
                                                {
                                                    Change_BalloonTipText(1);
                                                    iRightClickStatus = 1;      //next to right click
                                                }
                                                else 
                                                {
                                                    MessageBox.Show(togglesInfo[outsideToggle].mouse.ToUpper());
                                                }
                                                break;
                                            case 2:
                                                key = togglesInfo[outsideToggle].key1;
                                                if (key > 0)
                                                {
                                                    if ((key == 0x11 || key == 0x12) && togglesInfo[outsideToggle].key2 > 0)
                                                    {
                                                        InputCont.SendKey(key, 0);
                                                    }
                                                    else
                                                    {
                                                        InputCont.SendKey(key, 1);
                                                    }
                                                }
                                                key = togglesInfo[outsideToggle].key2;
                                                if (key > 0)
                                                {
                                                    InputCont.SendKey(key,1);
                                                }
                                                key = togglesInfo[outsideToggle].key3;
                                                if (key > 0)
                                                {
                                                    InputCont.SendKey(key,1);
                                                }
                                                key = togglesInfo[outsideToggle].key4;
                                                if (key > 0)
                                                {
                                                    InputCont.SendKey(key,1);
                                                }

                                                //if Ctrl , Alt  key release
                                                key = togglesInfo[outsideToggle].key1;
                                                if ((key == 0x11 || key == 0x12) && togglesInfo[outsideToggle].key2 > 0)
                                                {
                                                    InputCont.SendKey(key, 2);
                                                    //MessageBox.Show("release" + key.ToString("x"));
                                                }

                                                break;
                                            case 3:
                                                //MessageBox.Show(togglesInfo[outsideToggle].ps_path);
                                                if (togglesInfo[outsideToggle].ps_path.Trim().Length > 0)
                                                {
                                                    Process myProcess = new Process();
                                                    myProcess.StartInfo.FileName = togglesInfo[outsideToggle].ps_path;   //指定要執行的程式的路徑
                                                    try
                                                    {
                                                        myProcess.Start();
                                                    }
                                                    catch (Exception e)
                                                    {
                                                        
                                                        //MessageBox.Show(togglesInfo[outsideToggle].ps_path + ":\n" + e.Message);
                                                        Show_ErrorMessage(e.Message + "\n\n" + togglesInfo[outsideToggle].ps_path);
                                                    }
                                                }
                                                break;
                                        }
                                    }

                                }
                            }
                        } else if (CalibrationState.State > 0)
                            ContinueCalibration(position);
                    }
                    else if (CursorEnabled) //dragging
                    {
                        //show LoadingCricle at mouse cursor position
                        if (Math.Abs(clickPosition.X - lastWiiState.RawPosition.X) <= 5 && Math.Abs(clickPosition.Y - lastWiiState.RawPosition.Y) <= 5)
                        {
                            
                            if (clickCounter == 20 && !isShowLoadingCircle && iRightClickStatus < 1)
                            {
                                isShowLoadingCircle = true;
                                Show_LoadingCircle(warpedPos);
                            } else if(clickCounter > 180 && isShowLoadingCircle) {
                                Hide_LoadingCircle(); //add by gsyan
                                isShowLoadingCircle = false;
                            }
                        } else if(isShowLoadingCircle) {
                            Hide_LoadingCircle(); //add by gsyan
                            isShowLoadingCircle = false;
                        }


                        //if (((x != lastWiiState.RawPosition.X) || (y != lastWiiState.RawPosition.Y)) && cursorControl)
                        if (SmoothingEnabled)
                        {
                            PointF s = GetSmoothedCursor(SmoothingAmount);
                            InputCont.MoveMouse(s); //move the mouse to x,y
                        }
                        else
                            InputCont.MoveMouse(warpedPos); //move the mouse to x,y
                    }
                }
                else if (lastWiiState.Found && CursorEnabled)           //mouse up
                {

                    if (isShowLoadingCircle)
                    {
                        Hide_LoadingCircle(); //add by gsyan
                        isShowLoadingCircle = false;
                    }

                    //--------------------------------------
                    //right click
                    if (clickCounter > 60 && clickCounter < 180 && clickPosition.X >= 0 && clickPosition.Y >= 0 && Math.Abs(clickPosition.X - lastWiiState.RawPosition.X) <= 5 && Math.Abs(clickPosition.Y - lastWiiState.RawPosition.Y) <= 5)
                    {
                        InputCont.RightMouseDown();
                        InputCont.RightMouseUp();
                    }
                    else
                    {
                        if (iRightClickStatus < 1)
                        {
                            InputCont.LeftMouseUp();
                        } else {
                            if (iRightClickStatus++ == 2)
                            {
                                InputCont.RightMouseUp();
                                iRightClickStatus = 0;
                                
                                Change_BalloonTipText(0);
                            }
                            else if(iRightClickStatus>2) 
                            {
                                iRightClickStatus = 0;
                                
                                Change_BalloonTipText(0);
                            }
                        }
                    }
                    
                    clickCounter = 0;
                    clickPosition.X = -1;
                    clickPosition.Y = -1;

                    if (MoveCursor)
                        InputCont.MoveMouse(new PointF(0,0)); //move the mouse to x,y
                }

                if (!lastButtonState.A && bs.A)
                    Calibrate();

                lastButtonState = bs;
                lastWiiState = ws;

                if (StatusInfoChanged != null)
                    StatusInfoChanged(this, new EventArgs<WiimoteState>(args.WiimoteState));
            }
        }


        void ContinueCalibration(PointF position)
        {
            switch (CalibrationState.State)
            {
                case 1:
                    src[CalibrationState.State - 1] = position;
                    CalibrationState.TargetPoint.X = ScreenSize.Width - (int)(ScreenSize.Width * CALIBRATION_MARGIN);
                    CalibrationState.TargetPoint.Y = (int)(ScreenSize.Height * CALIBRATION_MARGIN);
                    dst[CalibrationState.State] = CalibrationState.TargetPoint;
                    CalibrationState.State++;
                    if (CalibrationStateChanged != null)
                        CalibrationStateChanged(this, new EventArgs<CalibrationStateInfo>(CalibrationState));
                    break;
                case 2:
                    src[CalibrationState.State - 1] = position;
                    CalibrationState.TargetPoint.X = (int)(ScreenSize.Width * CALIBRATION_MARGIN);
                    CalibrationState.TargetPoint.Y = ScreenSize.Height - (int)(ScreenSize.Height * CALIBRATION_MARGIN);
                    dst[CalibrationState.State] = CalibrationState.TargetPoint;
                    CalibrationState.State++;
                    if (CalibrationStateChanged != null)
                        CalibrationStateChanged(this, new EventArgs<CalibrationStateInfo>(CalibrationState));
                    break;
                case 3:
                    src[CalibrationState.State - 1] = position;
                    CalibrationState.TargetPoint.X = ScreenSize.Width - (int)(ScreenSize.Width * CALIBRATION_MARGIN);
                    CalibrationState.TargetPoint.Y = ScreenSize.Height - (int)(ScreenSize.Height * CALIBRATION_MARGIN);
                    dst[CalibrationState.State] = CalibrationState.TargetPoint;
                    CalibrationState.State++;
                    if (CalibrationStateChanged != null)
                        CalibrationStateChanged(this, new EventArgs<CalibrationStateInfo>(CalibrationState));
                    break;
                case 4:
                    src[CalibrationState.State - 1] = position;
                    warper.setDestination(dst[0], dst[1], dst[2], dst[3]);
                    warper.setSource(src[0], src[1], src[2], src[3]);
                    warper.computeWarp();
                    SaveConfig();
                    CalibrationState.State = 0;
                    if (CalibrationStateChanged != null)
                        CalibrationStateChanged(this, new EventArgs<CalibrationStateInfo>(CalibrationState));
                    CursorEnabled = PrevCursorEnabled;
                    break;
                default:
                    break;
            }
        }

        public void Calibrate()
        {
            PrevCursorEnabled = CursorEnabled;
            CursorEnabled = false;
            CalibrationState.TargetPoint.X = (int)(ScreenSize.Width * CALIBRATION_MARGIN);
            CalibrationState.TargetPoint.Y = (int)(ScreenSize.Height * CALIBRATION_MARGIN);
            dst[0] = CalibrationState.TargetPoint;

            //TODO: rather than src, dst being used directly while calibrating two temp arrays
            //should be used. so we still have the working values. (usefull to display a correct 
            //utilisation layout after a cancelling of a calibration

            CalibrationState.State = 1;
            if (CalibrationStateChanged != null)
                CalibrationStateChanged(this, new EventArgs<CalibrationStateInfo>(CalibrationState));
        }



        //add by gsyan
        public void Show_LoadingCircle(System.Drawing.PointF point) {
            ShowLoadingCircle(this, new EventArgs<System.Drawing.PointF>(point));
        }
        public void Hide_LoadingCircle()
        {
            HideLoadingCircle(this, EventArgs.Empty);
        }

        //add by gsyan
        public void Change_BalloonTipText(int status)
        {
            ChangeBalloonTipText(this, new EventArgs<int>(status));
        }
        
        //add by gsyan
        public void Show_ErrorMessage(String msg) {
            ShowErrorMessage(this,new EventArgs<String>(msg));
        }

        public void CalibrationCancel()
        {
            CalibrationState.State = 0;
            if (CalibrationStateChanged != null)
                CalibrationStateChanged(this, new EventArgs<CalibrationStateInfo>(CalibrationState));
            CursorEnabled = PrevCursorEnabled;
        }

        public void SaveConfig()
        {
            TextWriter tw = new StreamWriter("calibration.dat");
            for (int i = 0; i < 4; i++)
            {
                tw.WriteLine(src[i].X);
                tw.WriteLine(src[i].Y);
            }
            tw.Close();
        }

        public bool LoadConfig()
        {
            try
            {
                TextReader tr = new StreamReader("calibration.dat");
                for (int i = 0; i < 4; i++)
                {
                    src[i].X = float.Parse(tr.ReadLine());
                    src[i].Y = float.Parse(tr.ReadLine());
                }

                // close the stream
                tr.Close();

                warper.setDestination(new PointF(ScreenSize.Width * CALIBRATION_MARGIN,
                                        ScreenSize.Height * CALIBRATION_MARGIN),
                                      new PointF(ScreenSize.Width * (1.0f - CALIBRATION_MARGIN),
                                        ScreenSize.Height * CALIBRATION_MARGIN),
                                      new PointF(ScreenSize.Width * CALIBRATION_MARGIN,
                                        ScreenSize.Height * (1.0f - CALIBRATION_MARGIN)),
                                      new PointF(ScreenSize.Width * (1.0f - CALIBRATION_MARGIN),
                                        ScreenSize.Height * (1.0f - CALIBRATION_MARGIN)));
                warper.setSource(src[0], src[1], src[2], src[3]);

                warper.computeWarp(); 
                
                return true;
            }
            catch (System.IO.FileNotFoundException)
            {
                //no prexsting calibration
                return false;
            }
        }

    }
}
